(define-module (ordo action filesystem)
  #:use-module (ice-9 binary-ports)
  #:use-module (ice-9 match)
  #:use-module (rnrs bytevectors)
  #:use-module (srfi srfi-1) ; list utils
  #:use-module (srfi srfi-26) ; cut
  #:use-module (srfi srfi-71) ; extended let
  #:use-module ((srfi srfi-197) #:select (chain-when))
  #:use-module (ordo connection)
  #:use-module (ordo context)
  #:export (create-temporary-directory
            install-directory
            install-file
            fs:stat))

(define* (create-temporary-directory #:key tmpdir suffix template)
  (lambda (ctx)
    (must ctx "mktemp" (chain-when
                        '("--directory")
                        (tmpdir (append _ `("--tmpdir" tmpdir)))
                        (suffix (append _ `("--suffix" suffix)))
                        (template (append _ `(template)))))))

(define* (install-directory path #:key owner group mode)
  (lambda (ctx)
    (must ctx "install" (chain-when
                         '("--directory")
                         (owner (append _ `("--owner" ,owner)))
                         (group (append _ `("--group" ,group)))
                         (mode (append _ `("--mode" ,mode)))
                         (#t (append _ `(,path)))))))

;; Helper not intended for use outside of this module
(define (upload-tmp-file ctx)
  (lambda (input-port)
    (let ((tmp-path (first (must ctx "mktemp" `("-p" ,(context-scratch-dir ctx))))))
      (connection-call-with-output-file (context-connection ctx) tmp-path
                                        (lambda (output-port)
                                          (let loop ((data (get-bytevector-some input-port)))
                                            (unless (eof-object? data)
                                              (put-bytevector output-port data)
                                              (loop (get-bytevector-some input-port))))
                                          (close-port output-port)))
      tmp-path)))

;; Because we might need sudo to install the remote file, we first
;; upload the source to a temporary file.
(define* (install-file path #:key owner group mode content local-src remote-src backup?)
  (when (not (= 1 (length (filter identity (list content local-src remote-src)))))
    (error "exactly one of #:content, #:local-src, or #:remote-src is required"))
  (lambda (ctx)
    (let ((remote-src (cond
                       (remote-src remote-src)
                       (local-src (call-with-input-file local-src (upload-tmp-file ctx)))
                       ((string? content) (call-with-input-string content (upload-tmp-file ctx)))
                       ((bytevector? content) (call-with-input-bytevector content (upload-tmp-file ctx)))
                       (else (error "unsupported type for #:content")))))
      (must ctx "install" (chain-when
                           '()
                           (owner (append _ `("--owner" ,owner)))
                           (group (append _ `("--group" ,group)))
                           (mode (append _ `("--mode" ,mode)))
                           (backup? (append _ '("--backup" "numbered")))
                           (#t (append _ (list remote-src path)))))
      path)))

(define (fs:stat path)
  (define (parse-stat-result s)
    (match-let* (((file-type user group . rest) (string-split s #\:))
                 ((uid gid size mode atime mtime ctime) (map string->number rest)))
      `((file-type . ,file-type)
        (user . ,user)
        (group . ,group)
        (uid . ,uid)
        (gid . ,gid)
        (size . ,size)
        (mode . ,mode)
        (atime . ,atime)
        (mtime . ,mtime)
        (ctime . ,ctime))))
  (lambda (ctx)
    (let ((result rc (run ctx "stat" `("--format=%F:%U:%G:%u:%g:%s:#o%a:%X:%Y:%Z" ,path))))
      (cond
       ((zero? rc)  (parse-stat-result (first result)))
       ((string-contains? (first result) "No such file or directory") #f)
       (else (error (format #f "stat ~a: ~a" path (first result))))))))
